import 'dart:async';

import 'package:lori_sdk/data/database/interfaces/productDao.dart';
import 'package:lori_sdk/data/models/product/product.dart';
import 'package:lori_sdk/data/models/singleton/productSingleton.dart';
import 'package:path/path.dart';
import 'package:sqflite/sqflite.dart';

class MobileProductDao implements ProductDao {
  Database db;

  Future open() async {
    if (db == null || !db.isOpen) {
      var databasesPath = await getDatabasesPath();
      String path = join(databasesPath, 'loriProduct.db');
      db = await openDatabase(path, version: 1,
          onCreate: (Database db, int version) async {
        await db.execute('''
        create table $tableName (
          $columnId text,
          $columnName text,
          $columnCode text,
          $columnDescription text,
          $columnAdditionalInfo text,
          $columnImageUrl text,
          $columnEcommerceID text,
          $columnEcommercePrice number,
          $columnEcommercestock integer,
          $columnKioskId text,
          $columnKioskPrice number,
          $columnKioskstock integer,
          $columnDefaultPrice number,
          $columnSingleOldPrice number,
          $columnSinglePrice number,
          $columnMaxKioskInstallments integer,
          $columnKioskInstallmentPrice number,
          $columnMaxEcommerceInstallments integer,
          $columnEcommerceInstallmentPrice number,
          $columnMaxsingleInstallments integer,
          $columnSingleInstallmentPrice number,
          $columnIssinglePrice integer,
          $columnIsFavorite integer,
          $columnFavoriteCount integer,
          $columnIsNew integer,
          $columnIsAvailable integer,
          $columnLastTimestamp integer
        )
      ''');
      });
    }
  }

  Future<void> insertOrUpdate(Product product) async {
    var storedProducts = await get();
    Product readProduct;
    if (storedProducts != null) {
      readProduct =
          storedProducts.firstWhere((sp) => sp.id == product.id, orElse: () {
        return null;
      });
    }
    if (readProduct == null)
      return await insert(product);
    else
      return await update(product);
  }

  Future<void> insert(Product product) async {
    product.lastTimestamp = DateTime.now().millisecondsSinceEpoch;
    await open();
    await db.insert(tableName, product.toMap());
    return ProductSingleton().productsChanged(await get());
  }

  Future<void> update(Product product) async {
    product.lastTimestamp = DateTime.now().millisecondsSinceEpoch;
    await open();
    await db.update(tableName, product.toMap(),
        where: '$columnId = ?', whereArgs: [product.id]);
    return ProductSingleton().productsChanged(await get());
  }

  Future<List<Product>> get() async {
    List<Product> lastSeenProducts = [];
    await open();
    List<Map> maps = await db.query(tableName);
    if (maps != null && maps.length > 0) {
      for (var i = 0; i < maps.length; i++) {
        var product = Product();
        product.fromMap(maps[i]);
        lastSeenProducts.add(product);
      }
      lastSeenProducts
          .sort((p1, p2) => p2.lastTimestamp.compareTo(p1.lastTimestamp));
      return lastSeenProducts;
    }
    return null;
  }

  Future<void> deleteAll() async {
    await open();
    await db.delete(tableName);
    return ProductSingleton().productsChanged(null);
  }
}

ProductDao getProductDao() => MobileProductDao();
