import 'dart:async';
import 'dart:convert';

import 'package:lori_sdk/data/database/interfaces/userDao.dart';
import 'package:lori_sdk/data/models/singleton/userSingleton.dart';
import 'package:lori_sdk/data/models/user.dart';
import 'package:shared_preferences/shared_preferences.dart';

const int kVersion1 = 1;

class WebUserDao implements UserDao {
  static const _USER_KEY = 'x_user';

  SharedPreferences db;

  Future open() async {
    if (db == null) {
      db = await SharedPreferences.getInstance();
    }
  }

  Future<void> insertOrUpdate(User user) async {
    var storedUser = await get();
    if (storedUser == null) {
      return await insert(user);
    } else {
      return await update(user);
    }
  }

  Future<void> insert(User user) async {
    await open();
    await db.setString(_USER_KEY, jsonEncode(user.toJson()));
    return UserSingleton().userChanged(user);
  }

  Future<void> update(User user) async {
    await open();
    await db.setString(_USER_KEY, jsonEncode(user.toJson()));
    return UserSingleton().userChanged(user);
  }

  Future<User> get() async {
    await open();
    final userInJson = db.getString(_USER_KEY);
    return userInJson != null ? User.fromJson(jsonDecode(userInJson)) : null;
  }

  Future<void> deleteAll() async {
    await open();
    await db.setString(_USER_KEY, null);
    return;
  }
}

UserDao getUserDao() => WebUserDao();
