import 'package:json_annotation/json_annotation.dart';

part 'authResponse.g.dart';

final String tableName = 'auth_response';
final String columnToken = 'access_token';
final String columnRefreshToken = 'refresh_token';
final String columnExpiresIn = 'expires_in';
final String columnTokenType = 'token_type';
final String columnResponseType = 'response_type';
final String columnResponseCreateDate = 'response_create_date';

@JsonSerializable()
class AuthResponse {
  static const int responseUser = 1;
  static const int responseClient = 2;

  AuthResponse(
      this.accessToken, this.refreshToken, this.expiresIn, this.tokenType);

  @JsonKey(name: 'access_token')
  String accessToken;
  @JsonKey(name: 'refresh_token')
  String refreshToken;
  @JsonKey(name: 'expires_in')
  int expiresIn;
  @JsonKey(name: 'token_type')
  String tokenType;
  int responseType;
  int responseCreatedDate;

  bool isTokenValid() {
    if (responseCreatedDate == null) return true;
    var tokenDate = DateTime.fromMillisecondsSinceEpoch(responseCreatedDate);
    tokenDate.add(Duration(seconds: expiresIn - 30));
    return DateTime.now().isBefore(tokenDate);
  }

  factory AuthResponse.fromJson(Map<String, dynamic> json) =>
      _$AuthResponseFromJson(json);

  Map<String, dynamic> toJson() => _$AuthResponseToJson(this);

  // Helpers method to support SQLite
  Map<String, dynamic> toMap() {
    var map = <String, dynamic>{
      columnToken: accessToken,
      columnRefreshToken: refreshToken,
      columnExpiresIn: expiresIn,
      columnTokenType: tokenType,
      columnResponseType: responseType,
      columnResponseCreateDate: responseCreatedDate,
    };
    return map;
  }

  void fromMap(Map<String, dynamic> map) {
    accessToken = map[columnToken];
    refreshToken = map[columnRefreshToken];
    expiresIn = map[columnExpiresIn];
    tokenType = map[columnTokenType];
    responseType = map[columnResponseType];
    responseCreatedDate = map[columnResponseCreateDate];
  }
}
