import 'package:json_annotation/json_annotation.dart';
import 'package:lori_sdk/ext/extensions.dart';

part 'voucher.g.dart';

@JsonSerializable()
class Voucher {
  Voucher();

  String id;
  int usageLimit;
  String code;

  @JsonKey(defaultValue: false)
  bool hasNoFeeInstallments;

  @JsonKey(defaultValue: false)
  bool freeShipping;

  double discountPercentage;
  double discountValue;

  String description() {
    return code.toUpperCase() + ' - ' + _valueDescription();
  }

  String _valueDescription() {
    if (hasNoFeeInstallments) {
      return 'Parcelamento sem juros';
    } else if (freeShipping) {
      return 'Frete Grátis';
    } else if (discountPercentage != null) {
      return discountPercentage.toStringAsFixed(0) + '%';
    } else {
      //Setting discountValue as future-proof `default` description
      return (discountValue ?? 0.0).formatDoubleToCurrency();
    }
  }

//  double discount(double totalValue) {
//    return discountValue != null && discountValue > 0 ? discountValue : ((discountPercentage / 100) * totalValue);
//  }

  factory Voucher.fromJson(Map<String, dynamic> json) =>
      _$VoucherFromJson(json);

  Map<String, dynamic> toJson() => _$VoucherToJson(this);
}
