import 'package:json_annotation/json_annotation.dart';

part 'notification.g.dart';

final String tableName = 'notification';
final String columnId = 'id';
final String columnNotificationId = 'notificationId';
final String columnTitle = 'title';
final String columnMessage = 'message';
final String columnImageUrl = 'image_url';
final String columnIsRead = 'is_read';
final String columnAction = 'action';
final String columnObjectReferenceId = 'object_reference_id';
final String columnOrderId = 'order_id';
final String columnExternalId = 'external_id';

@JsonSerializable()
class Notification {
  static const NOTIFICATION_TITLE = 'title';
  static const NOTIFICATION_BODY = 'body';
  static const NOTIFICATION_IMAGE_URL = 'imageUrl';
  static const NOTIFICATION_ACTION = 'action';
  static const NOTIFICATION_ORDER_ID = 'orderId';
  static const NOTIFICATION_EXTERNAL_ID = 'externalId';

  static const String ACTION_PROMOTION = "open_promotion";
  static const String ACTION_NEWS = "open_promotion";
  static const String ACTION_TAG = "open_tag";
  static const String ACTION_PRODUCT = "open_product";
  static const String ACTION_CATEGORY = "open_category";
  static const String ACTION_SUBCATEGORY = "open_subcategory";
  static const String ACTION_ORDER_PAYMENT_STATUS = "order_status_payment";
  static const String ACTION_ORDER_STATUS = "order_status";

  static const String ACTION_PROMOTION_OLD = "action_promotion";
  static const String ACTION_PRODUCT_OLD = "action_open_product";
  static const String ACTION_CATEGORY_OLD = "action_open_category";

  Notification(
      this.id,
      this.title,
      this.message,
      this.notificationId,
      this.imageUrl,
      this.isRead,
      this.action,
      this.objectReferenceId,
      this.orderId,
      this.externalId);

  Notification.empty();

  int id;
  String title;
  String message;
  String notificationId;
  String imageUrl;
  bool isRead;
  String action;
  String objectReferenceId;
  String orderId;
  String externalId;

  Map<dynamic, dynamic> transformed;

  static Map<dynamic, dynamic> transform(Map<dynamic, dynamic> dataFromPush) {
    return {
      'notification': {
        'title': dataFromPush['title'],
        'body': dataFromPush['body'],
      },
      'data': {
        'notificationId': dataFromPush['notificationId'],
        'imageUrl': dataFromPush['imageUrl'],
        'action': dataFromPush['action'],
        'referenceId': dataFromPush['referenceId'],
        'orderId': dataFromPush['orderId'],
        'externalId': dataFromPush['externalId']
      }
    };
  }

  static Notification parse(Map<dynamic, dynamic> notificationData,
      Map<dynamic, dynamic> messageData) {
    return Notification(
        0,
        notificationData['title'],
        notificationData['body'],
        messageData['notificationId'],
        messageData['imageUrl'],
        false,
        messageData['action'],
        messageData['referenceId'],
        messageData['orderId'],
        messageData['externalId']);
  }

  factory Notification.fromJson(Map<String, dynamic> json) =>
      _$NotificationFromJson(json);
  Map<String, dynamic> toJson() => _$NotificationToJson(this);

  // Helpers method to support SQLite
  Map<String, dynamic> toMap() {
    var map = <String, dynamic>{
      columnId: id,
      columnTitle: title,
      columnMessage: message,
      columnNotificationId: notificationId,
      columnImageUrl: imageUrl,
      columnIsRead: isRead ? 1 : 0,
      columnAction: action,
      columnObjectReferenceId: objectReferenceId,
      columnOrderId: orderId,
      columnExternalId: externalId,
    };
    return map;
  }

  void fromMap(Map<String, dynamic> map) {
    id = map[columnId];
    title = map[columnTitle];
    message = map[columnMessage];
    notificationId = map[columnNotificationId];
    imageUrl = map[columnImageUrl];
    isRead = map[columnId] == 1 ? true : false;
    action = map[columnAction];
    objectReferenceId = map[columnObjectReferenceId];
    orderId = map[columnOrderId];
    externalId = map[columnExternalId];
  }
}
