import 'package:json_annotation/json_annotation.dart';
import 'package:lori_sdk/data/models/product/productImage.dart';
import 'package:lori_sdk/data/models/product/skuAttribute.dart';

part 'product.g.dart';

const ORDER_RELEVANCE = "Relevance";
const ORDER_HIGHER = "HigherPrice";
const ORDER_LOWEST = "LowestPrice";
const ORDER_FAVORED = "MostFavored";
const ORDER_SELLERS = "BestSellers";

final String tableName = 'product';
final String columnId = 'id';
final String columnName = 'name';
final String columnCode = 'code';
final String columnDescription = 'description';
final String columnAdditionalInfo = 'additional_info';
final String columnImageUrl = 'image_url';
final String columnEcommerceID = 'ecommerce_id';
final String columnEcommercePrice = 'ecommerce_price';
final String columnEcommercestock = 'ecommerce_stock';
final String columnKioskId = 'kiosk_id';
final String columnKioskPrice = 'kiosk_price';
final String columnKioskstock = 'kiosk_stock';
final String columnDefaultPrice = 'default_price';
final String columnSingleOldPrice = 'single_old_price';
final String columnSinglePrice = 'single_price';
final String columnMaxKioskInstallments = 'max_kiosk_installments';
final String columnKioskInstallmentPrice = 'kiosk_installment_price';
final String columnMaxEcommerceInstallments = 'max_ecommerce_installments';
final String columnEcommerceInstallmentPrice = 'ecommerce_installment_price';
final String columnMaxsingleInstallments = 'max_single_installments';
final String columnSingleInstallmentPrice = 'single_installment_price';
final String columnIssinglePrice = 'is_single_price';
final String columnIsFavorite = 'is_favorite';
final String columnFavoriteCount = 'favorite_count';
final String columnIsNew = 'is_new';
final String columnIsAvailable = 'is_available';
final String columnLastTimestamp = 'lastTimestamp';

@JsonSerializable()
class Product {
  Product();

  String id;
  String name;
  String code;
  String description;
  String additionalInfo;
  String imageUrl;

  String ecommerceId;
  double ecommercePrice;
  int ecommerceStock;

  String kioskId;
  double kioskPrice;
  int kioskStock;

  double defaultPrice;
  double singleOldPrice;
  double singlePrice;

  int maxKioskInstallments;
  double kioskInstallmentPrice;
  int maxEcommerceInstallments;
  double ecommerceInstallmentPrice;
  int maxSingleInstallments;
  double singleInstallmentPrice;
  bool isSinglePrice;

  List<Sku> skus;
  List<ProductImage> images;

  bool isFavorite;
  int favoriteCount;
  bool isNew;
  bool isAvailable;
  int lastTimestamp;

  double getOldPrice() {
    if (isSinglePrice) {
      return singleOldPrice;
    }
    else {
      //TODO: Add this to Kiosk/Ecommerce Price if we ever use it again
      return singleOldPrice;
    }
  }

  double getProductPrice() {
    if (isSinglePrice)
      return singlePrice;
    else {
      if (this.kioskPrice <= this.ecommercePrice)
        return this.kioskPrice;
      else
        return this.ecommercePrice;
    }
  }

  double getInstallmentPrice() {
    if (isSinglePrice)
      return this.singleInstallmentPrice;
    else {
      if (this.kioskInstallmentPrice <= this.ecommerceInstallmentPrice)
        return this.kioskInstallmentPrice;
      else
        return this.ecommerceInstallmentPrice;
    }
  }

  int getMaxInstallments() {
    if (isSinglePrice)
      return this.maxSingleInstallments;
    else {
      if (this.maxKioskInstallments <= this.maxEcommerceInstallments)
        return this.maxKioskInstallments;
      else
        return this.maxEcommerceInstallments;
    }
  }

  factory Product.fromJson(Map<String, dynamic> json) =>
      _$ProductFromJson(json);

  Map<String, dynamic> toJson() => _$ProductToJson(this);

  Map<String, dynamic> toMap() {
    var map = <String, dynamic>{
      columnId: id,
      columnName: name,
      columnCode: code,
      columnDescription: description,
      columnAdditionalInfo: additionalInfo,
      columnImageUrl: imageUrl,
      columnEcommerceID: ecommerceId,
      columnEcommercePrice: ecommercePrice?.toDouble(),
      columnEcommercestock: ecommerceStock,
      columnKioskId: kioskId,
      columnKioskPrice: kioskPrice?.toDouble(),
      columnKioskstock: kioskStock,
      columnDefaultPrice: defaultPrice?.toDouble(),
      columnSingleOldPrice: singleOldPrice?.toDouble(),
      columnSinglePrice: singlePrice?.toDouble(),
      columnMaxKioskInstallments: maxKioskInstallments,
      columnKioskInstallmentPrice: kioskInstallmentPrice?.toDouble(),
      columnMaxEcommerceInstallments: maxEcommerceInstallments,
      columnEcommerceInstallmentPrice: ecommerceInstallmentPrice?.toDouble(),
      columnMaxsingleInstallments: maxSingleInstallments,
      columnSingleInstallmentPrice: singleInstallmentPrice?.toDouble(),
      columnIssinglePrice: isSinglePrice ? 1 : 0,
      columnIsFavorite: isFavorite ? 1 : 0,
      columnFavoriteCount: favoriteCount,
      columnIsNew: isNew ? 1 : 0,
      columnIsAvailable: isAvailable ? 1 : 0,
      columnLastTimestamp: lastTimestamp,
    };
    return map;
  }

  void fromMap(Map<String, dynamic> map) {
    id = map[columnId];
    name = map[columnName];
    code = map[columnCode];
    description = map[columnDescription];
    additionalInfo = map[columnAdditionalInfo];
    imageUrl = map[columnImageUrl];
    ecommerceId = map[columnEcommerceID];
    ecommercePrice = map[columnEcommercePrice]?.toDouble();
    ecommerceStock = map[columnEcommercestock];
    kioskId = map[columnKioskId];
    kioskPrice = map[columnKioskPrice]?.toDouble();
    kioskStock = map[columnKioskstock];
    defaultPrice = map[columnDefaultPrice]?.toDouble();
    singleOldPrice = map[columnSingleOldPrice]?.toDouble();
    singlePrice = map[columnSinglePrice]?.toDouble();
    maxKioskInstallments = map[columnMaxKioskInstallments];
    kioskInstallmentPrice = map[columnKioskInstallmentPrice]?.toDouble();
    maxEcommerceInstallments = map[columnMaxEcommerceInstallments];
    ecommerceInstallmentPrice =
        map[columnEcommerceInstallmentPrice]?.toDouble();
    maxSingleInstallments = map[columnMaxsingleInstallments];
    singleInstallmentPrice = map[columnSingleInstallmentPrice]?.toDouble();
    isSinglePrice =
        map[columnIssinglePrice] != null && map[columnIssinglePrice] == 1
            ? true
            : false;
    isFavorite = map[columnIsFavorite] != null && map[columnIsFavorite] == 1
        ? true
        : false;
    favoriteCount = map[columnFavoriteCount];
    isNew = map[columnIsNew] != null && map[columnIsNew] == 1 ? true : false;
    isAvailable = map[columnIsAvailable] != null && map[columnIsAvailable] == 1
        ? true
        : false;
    lastTimestamp = map[columnLastTimestamp];
  }
}

@JsonSerializable()
class Sku {
  Sku();

  String id;
  String code;
  String name;
  String imageUrl;
  String selectionImageUrl;
  String color;

  String productSku;
  String productCode;
  Product product;

  List<ProductImage> images;
  List<SkuSpecificAttribute> attributes;

  bool isAvailable;

  String correctName() {
    return name == null ? code : name;
  }

  factory Sku.fromJson(Map<String, dynamic> json) => _$SkuFromJson(json);

  Map<String, dynamic> toJson() => _$SkuToJson(this);
}